#include <fcntl.h>
#include <stdint.h>
#include <stdio.h>
#include <string.h>
#include <termios.h>
#include <time.h>
#include <unistd.h>

int serial_file=-1;

double get_time()
{
   // high res time
   struct timespec ts;
   clock_gettime(CLOCK_REALTIME, &ts);
   return (double)ts.tv_sec+(1.0e-9*(double)(ts.tv_nsec));
}

int serial_init(const char *serial_device)
{
   // open the serial device (not a controlling terminal)
   serial_file=open(serial_device, O_RDWR | O_NOCTTY);

   if (serial_file<0)
   {
      // open failed
      return 0;
   }

   // start with clean new settings
   struct termios new_serial_settings;
   memset(&new_serial_settings, 0, sizeof(new_serial_settings));

   // 9600 baud 8N1, read enabled, no modem control signals needed
   new_serial_settings.c_cflag=(B9600 | CS8 | CLOCAL | CREAD);

   // raw output
   new_serial_settings.c_oflag=0;

   // turn off canonical input
   new_serial_settings.c_lflag=0;

   // timeout of 400 ms
   new_serial_settings.c_cc[VMIN]  = 0x0;
   new_serial_settings.c_cc[VTIME] = 20;

   // flush the port
   tcflush(serial_file, TCIOFLUSH);

   // apply new settings to the serial port
   tcsetattr(serial_file, TCSANOW, &new_serial_settings);

   return 1;
}

void serial_reset()
{
   //printf("# resetting serial port");
   if (serial_file>=0)
   {
      // flush the port
      tcflush(serial_file, TCIOFLUSH);

      sleep(4);

      // start with clean new settings
      struct termios new_serial_settings;
      memset(&new_serial_settings, 0, sizeof(new_serial_settings));

      // 9600 baud 8N1, read enabled, no modem control signals needed
      new_serial_settings.c_cflag=(B9600 | CS8 | CLOCAL | CREAD);

      // raw output
      new_serial_settings.c_oflag=0;

      // turn off canonical input
      new_serial_settings.c_lflag=0;

      // timeout of 400 ms
      new_serial_settings.c_cc[VMIN]  = 0x0;
      new_serial_settings.c_cc[VTIME] = 20;

      // flush the port
      tcflush(serial_file, TCIOFLUSH);

      // apply new settings to the serial port
      tcsetattr(serial_file, TCSANOW, &new_serial_settings);
   }
}

// verify 4 byte header of the reply, return zero if checksum fails
uint8_t check_header(const uint8_t* buffer, int x)
{
   uint8_t cs=0;
   while(x--)
   {
      cs^=*(buffer++);
   }
   return(!cs);
}

// verify x byte data section of the reply, return zero if checksum fails
uint8_t check_payload(const uint8_t* buffer, int x)
{
   uint8_t cs=0;
   if (x>4)
   {
      x-=4;
      buffer+=4;
   }

   while(x--)
   {
      cs^=*(buffer++);
   }
   return(!cs);
}


// wait for reply_size bytes, return number of bytes placed in rx_buffer
int read_reply(uint8_t *rx_buffer, int reply_size)
{
   int rx_count=0;
   while (rx_count<reply_size)
   {
      int x=read(serial_file, rx_buffer, reply_size-rx_count);
      if (0==x)
      {
         // timeout!
         return(0);
      }
      rx_count+=x;
      rx_buffer+=x;
   }
   return(rx_count);
}

// transmit a 4 byte command to the FE-5680A, wait for a reply, verify header and data reply
// return value if the number of bytes placed in the rx_buffer
int tx_command(uint8_t command, uint8_t reply_length, uint8_t *rx_buffer)
{
   uint8_t tx_buffer[4];
   tx_buffer[0]=command;
   tx_buffer[1]=4;
   tx_buffer[2]=0;
   tx_buffer[3]=tx_buffer[0] ^ tx_buffer[1] ^ tx_buffer[2];    // calc checksum
   write(serial_file, tx_buffer, sizeof(tx_buffer));

   if (reply_length==read_reply(rx_buffer, reply_length))
   {
      // correct length reply received.
      // are the checksums correct?
      if ((check_header(rx_buffer, 4)) && (check_payload(rx_buffer, reply_length)))
      {
         return(reply_length);
      }
   }
   // either the command timed out, or the checksum was wrong
   // reset the serial port
   serial_reset();
   return(0);
}

// transmit a 4 byte command to the FE-5680A, wait for a reply, verify header and data reply
// return value if the number of bytes placed in the rx_buffer
int tx_command2(uint8_t command, uint8_t *rx_buffer)
{
   uint8_t tx_buffer[4];
   uint8_t reply_length;

   tx_buffer[0]=command;
   tx_buffer[1]=4;
   tx_buffer[2]=0;
   tx_buffer[3]=tx_buffer[0] ^ tx_buffer[1] ^ tx_buffer[2];    // calc checksum
   write(serial_file, tx_buffer, sizeof(tx_buffer));

   if (0x04==read_reply(rx_buffer, 4))
   {
      //printf("hdr ");
      // header received, check length and receive the data
      if (check_header(rx_buffer, 4))
      {
         //printf("ok ");
         // second byte is data payload length
         reply_length=rx_buffer[1]-4;
         rx_buffer+=4;

         if (reply_length==read_reply(rx_buffer, reply_length))
         {
            //printf("data ");
            if (check_payload(rx_buffer-4, reply_length+4))
            {
               printf("cs ok ");
            }
            return(reply_length+4);
         }
      }
   }
   serial_reset();
   return(0);
}

// use command 0x5A to read the 4 channel 12 bit ADC
// we think channel 1 is the pot, 2 is a rail, 3 is temperature related, and 4 is vcxo related
void read_adc()
{
   // Reply to command 0x5A, 0x0D bytes: [5A] [0D] [00] [57] [00] [00] [40] [08] [9E] [07] [00] [00] [D1]
   int i;
   uint8_t rx_buffer[256];
   uint16_t adc;

   if (tx_command(0x5a, 0x0d, rx_buffer))
   {
      for (i=0; i<4; i++)
      {
         adc=rx_buffer[(i*2)+4] | (rx_buffer[(i*2)+5] << 8);
         printf("  %4d  ", adc);
      }
   }

}

// use command 0x53 to read a single 16 bit reply
// the data appears to be temperature related
void test_53_command()
{
   uint8_t rx_buffer[256];
   uint16_t adc;

   if (tx_command(0x53, 0x07, rx_buffer))
   {
      adc=rx_buffer[4] | (rx_buffer[5] << 8);
      printf("%4d ", adc);
   }
}





// transmit a 4 byte command to the FE-5680A, wait for a reply, verify header and data reply
// return value is the number of bytes placed in the rx_buffer
int tx_command_3(uint8_t command, uint8_t *rx_buffer)
{
   uint8_t tx_buffer[4];
   uint8_t reply_length;

   tx_buffer[0]=command;
   tx_buffer[1]=4;
   tx_buffer[2]=0;
   tx_buffer[3]=tx_buffer[0] ^ tx_buffer[1] ^ tx_buffer[2];    // calc checksum
   write(serial_file, tx_buffer, sizeof(tx_buffer));

   if (0x04==read_reply(rx_buffer, 4))
   {
      // header received, check length and receive the data
      if (check_header(rx_buffer, 4))
      {
         // second byte is data payload length
         reply_length=rx_buffer[1]-4;
         rx_buffer+=4;

         if (reply_length==read_reply(rx_buffer, reply_length))
         {
            if (!check_payload(rx_buffer-4, reply_length+4))
            {
               // checksum was wrong, so note it (but don't reset serial port)
               printf("(CS bad!) ");
            }
            return(reply_length+4);
         }
      }
   }
   serial_reset();
   return(0);
}




// use command 0x22 to read 4 16 bit values
// They appear to be the two DDS tuning words
void test_22_command()
{
   int reply_length;
   int i;
   uint32_t dds_1;
   uint32_t dds_2;
   static uint32_t last_word=0;
   uint8_t rx_buffer[256];
   uint8_t command=0x22;

   reply_length=tx_command_3(command, rx_buffer);

   // skip checksum
   reply_length-=1;

   int index=4;
   for (i=0; i<4; ++i)
   {
      dds_1=(dds_1 << 8);
      dds_1|=rx_buffer[index];
      ++index;
   }

   for (i=0; i<4; ++i)
   {
      dds_2=(dds_2 << 8);
      dds_2|=rx_buffer[index];
      ++index;
   }

   printf("%lu %lu %ld ", dds_1, dds_2, (int32_t)(dds_2-last_word));

   last_word=dds_2;
}

// use command 0x2d to read a 32 bit signed value (user tuning value)
void test_2d_command()
{
   int reply_length;
   int i;
   int32_t dds;
   uint8_t rx_buffer[256];
   uint8_t command=0x2d;

   reply_length=tx_command_3(command, rx_buffer);

   // skip checksum
   reply_length-=1;

   int index=4;
   for (i=0; i<4; ++i)
   {
      dds=(dds << 8);
      dds|=rx_buffer[index];
      ++index;
   }

   printf("%ld ", dds);
}

// use command 0x29 to read 4 8 bit values
void test_29_command()
{
   int i;
   uint8_t rx_buffer[256];

   if (tx_command(0x29, 0x09, rx_buffer))
   {
      for (i=0; i<4; i++)
      {
         printf("%3d ", rx_buffer[4+i]);
      }
   }
}

// use command 0x2b to read x 8 bit values
void test_2b_command()
{
   int i;
   uint8_t rx_buffer[256];

   i=tx_command2(0x2b, rx_buffer);
   if (i)
   {
      while (i--)
      {
         printf("%3d ", rx_buffer[i+4]);
      }
   }
}

// use command 0x47 to read 4 8 bit values
void test_47_command()
{
   int i;
   uint8_t rx_buffer[256];

   i=tx_command2(0x47, rx_buffer);
   if (i)
   {
      while (i--)
      {
         printf("%3d ", rx_buffer[i+4]);
      }
   }
}

// use command 0x5a to read 4 8 bit values
void test_5a_command()
{
   int i;
   uint8_t rx_buffer[256];

   i=tx_command2(0x5a, rx_buffer);
   if (i)
   {
      while (i--)
      {
         printf("%3d ", rx_buffer[i+4]);
      }
   }
}

// use command 0x61 to read 4 8 bit values
void test_61_command()
{
   int i;
   uint8_t rx_buffer[256];

   i=tx_command2(0x61, rx_buffer);
   if (i)
   {
      while (i--)
      {
         printf("%3d ", rx_buffer[i+4]);
      }
   }
}

// use command 0x65 to read 4 8 bit values
void test_65_command()
{
   int i;
   uint8_t rx_buffer[256];

   i=tx_command2(0x65, rx_buffer);
   if (i)
   {
      while (i--)
      {
         printf("%3d ", rx_buffer[i+4]);
      }
   }
}

// use command 0x67 to read 4 8 bit values
void test_67_command()
{
   int i;
   uint8_t rx_buffer[256];

   i=tx_command2(0x67, rx_buffer);
   if (i)
   {
      while (i--)
      {
         printf("%3d ", rx_buffer[i+4]);
      }
   }
}


void test_command(uint8_t command)
{
   int i, j;
   uint8_t rx_buffer[256];

   printf("\n#\n# Cmd 0x%2.2X ", command);
   i=tx_command2(command, rx_buffer);
   printf("0x%2.2X byte reply: ", i);

   for (j=0; j<i; ++j)
   {
      printf("[%2.2X] ", rx_buffer[j]);
   }

   printf("\n# ASCII (.): ");
   for (j=0; j<i; ++j)
   {
      if ((rx_buffer[j]>=' ') && (rx_buffer[j]<='~'))
      {
         printf("%c", rx_buffer[j]);
      }
      else
      {
         printf(".");
      }
   }

   printf("\n# ASCII ( ): ");
   for (j=0; j<i; ++j)
   {
      if ((rx_buffer[j]>=' ') && (rx_buffer[j]<='~'))
      {
         printf("%c", rx_buffer[j]);
      }
      else
      {
         printf(" ");
      }
   }
}


int main(int argc, const char *argv[])
{
   double start_time;

   if (argc<2)
   {
      printf("\nUsage: %s comport\n", argv[0]);
      return(1);
   }

   if (0==serial_init(argv[1]))
   {
      printf("\nCouldn't open port %s\n", argv[1]);
      return(1);
   }

   start_time=get_time();
   time_t now=time(0);
   printf("# Logging started at %s", ctime(&now));
   printf("# Elapsed (s), rail?, trimpot, temperature, vcxo, dds 1, dds 2, delta dds2, user tuning");

   for(;;)
   {
      printf("\n     %9.3f     ", get_time()-start_time);
      read_adc();
      test_22_command();
      test_2d_command();
      fflush(0);
   }
}
